// Spannungsueberwachung im Arduino UNO / MEGA / MEGA2560
// (c) 2017 softwareentwicklung-als-prozess.de

#include "mqttclient.h"
#include "getfunctions.h"

#define Set_REF_IS_VCC()     (ADMUX  |= (1<<REFS0))
# if defined(__AVR_ATmega328P__)
#define Set_INTREF_INPUT()   (ADMUX  |= (1<<MUX3) | (1<<MUX2) | (1<<MUX1))
#endif
#if defined(__AVR_ATmega2560__) // Der 2560 hat ein MUX bit mehr fuer die 1.1Volt
#define Set_INTREF_INPUT()   (ADMUX  |= (1<<MUX4) | (1<<MUX3) | (1<<MUX2) | (1<<MUX1))
#endif
#define ADC_ENABLE()         (ADCSRA |= (1<<ADEN))
#define ADC_START()          (ADCSRA |= (1<<ADSC))
#define ADC_BUSY()           bitRead(ADCSRA, ADSC)
#define MV_BIT               1125300UL   // 1100mV * 1023 = 1125300
#define NORMAL_LVL           3000UL
#define EINGESCHR_LVL        2000UL
#define BO_RESUME_LVL        1900UL
#define BO_WARN_LVL          1800UL

int  led = (int)13;              // Pin der LED
unsigned int  adc_low, adc_high; // Zwischenspeicher für die Ergebnisse des ADC
unsigned long adc_result;        // Gesamtergebnis der Messung des ADC
unsigned long vcc;               // Versorgungsspannung
boolean brownoutwarning = (boolean)false;
Betriebsart betrieb = None;

void setup() 
{    
  pinMode(led, OUTPUT);  

  Serial.begin(9600);   // USB Ausgabe zum Debuggen
  Serial.println(F("[Spannungsueberwachung]"));
  // Ethernet / MQTT
  if (MQTTSetup() == false)
  {
    Serial.println("Ethernet Fehler - DHCP nicht erfolgreich");
  }
  else
  {
    Serial.println("- Ethernet aktiviert -");
    PrintIPAddress();
  }
  // AD Converter
  Set_REF_IS_VCC();     // VCC als Referenzspannung für den AD-Wandler
  Set_INTREF_INPUT();   // 1.1V Referenzspannung als Eingang für ADC 
  delay(10);            // warten bis sich die Referenzspannung eingestellt hat
  ADC_ENABLE();         // ADC aktivieren
}

void loop() 
{    
  ADC_START();          // ADC Messung starten
  while (ADC_BUSY())    // warten bis Messung beendet ist
  {
    ;
  }
  adc_low  = ADCL;
  adc_high = ADCH;

  adc_result = Get_Result(adc_high, adc_low); // Gesamtergebniss der ADC-Messung 
  vcc = MV_BIT / adc_result;                  // Versorgungsspannung in mV berechnen 

  Serial.print(F("Spannung = "));
  Serial.print(vcc, DEC);
  Serial.println(F(" mV"));

  if (vcc >= NORMAL_LVL)
  {
    // Normalbetrieb
    digitalWrite(led, LOW);  //schalte LED aus
    betrieb = Normal;
  }
  else
  {
    if (vcc >= EINGESCHR_LVL)
    {
      // Eingeschraenkter Betrieb
      betrieb = Eingeschraenkt;
    }
    else
    {
      if ((vcc > BO_RESUME_LVL) && ((boolean)true == brownoutwarning))
      {
        // Resume Brown Out
        brownoutwarning = (boolean)false;
        betrieb = BO_Resume;
      }
      else
      {
        if (vcc <= BO_WARN_LVL)
        {
          // Brown Out, soweit noch erkennbar
          digitalWrite(led, HIGH);  //schalte LED an
          brownoutwarning = (boolean)true;
          betrieb = BO_Warnung;
        }
      }
    }
  }
  Serial.print(F("Betriebsart = "));
  switch(betrieb)
  {
    case Normal:
      Serial.println(F("Normaler Betrieb"));
      break;
    case Eingeschraenkt:
      Serial.println(F("Eingeschraenkter Betrieb"));
      break;
    case BO_Resume:
      Serial.println(F("Eingeschraenkter Betrieb / Brown Out ueberstanden"));
      break;
    case BO_Warnung:
      Serial.println(F("Brown Out Warnung!"));
      break;    
  }
  MQTTConnectAndPublish();                     // Messwerte zum MQTT Server senden
  delay(20000);
}

unsigned long Get_Result(unsigned int high, unsigned int low)
{
  unsigned long ret = 0UL;
  
  ret = (unsigned long)((high << 8) | low);
  return(ret);
}

void PrintIPAddress(void)
{
  Serial.print("Die IP Adresse ist: ");
  for (byte thisbyte = 0; thisbyte < 4; thisbyte++)
  {
    Serial.print(MQTTGetIPAddress(thisbyte),DEC);
    if (thisbyte < 3)
    {
      Serial.print(".");
    }
  }
  Serial.println();
}

unsigned long Get_Spannung(void)
{
  return(vcc);
}

Betriebsart Get_Status(void)
{
  return(betrieb);
}

